# -*- perl -*-
#
# Test::MemMonitor::Backends::Linux by Daniel Berrange <berrange@redhat.com>
#
# Copyright (C) 2006 Red Hat <http://redhat.com/>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

=pod

=head1 NAME

Test::MemMonitor::Backends::Linux - backend for collecting process memory data on Linux

=head1 SYNOPSIS

  use Test::MemMonitor:Backends::Linux;



=head1 DESCRIPTION

This module provides a backend for collecting process memory data on Linux
based systems.

=head1 METHODS

=over 4

=cut

package Test::MemMonitor::Backends::Linux;

use strict;
use warnings;

use File::Spec::Functions qw(rootdir catfile catdir splitpath);
use IO::File;

use Test::MemMonitor::Grouping;
use Test::MemMonitor::Mapping;
use Test::MemMonitor::Snapshot;

sub snapshot {
    my $pid = shift;
    
    my $snapshot = &load_statm($pid);
    
    my $current = 0;
    my $total = 0;
    my $free = 0;
    
    my %groupings;
    my @mappings;
    foreach my $mapping (&load_mappings($pid)) {
	if ($mapping->start_address != $current) {
	    my $freemap = Test::MemMonitor::Mapping->new();
	    $freemap->start_address($current);
	    $freemap->end_address($mapping->start_address);
	    $freemap->readable(0);
	    $freemap->writable(0);
	    $freemap->executable(0);
	    $freemap->shared(0);
	    $freemap->private(0);
	    $freemap->offset(0);
	    $freemap->majordev(0);
	    $freemap->minordev(0);
	    $freemap->inode(0);
	    $freemap->path("[free]");
	    $freemap->deleted(0);
	    $freemap->shared_clean(0);
	    $freemap->shared_dirty(0);
	    $freemap->private_clean(0);
	    $freemap->private_dirty(0);

	    push @mappings, $freemap;	    
	    $free += $freemap->size;
	    
	    unless (exists $groupings{$freemap->path}) {
		$groupings{$freemap->path} = Test::MemMonitor::Grouping->new();
		$groupings{$freemap->path}->size(0);
		$groupings{$freemap->path}->name($freemap->path);
	    }
	    
	    $groupings{$freemap->path}->add_mapping($freemap);
	}
	
	push @mappings, $mapping;
	$total += $mapping->size;
	
	my $key;
	if ($mapping->path =~ m,^/,) {
	    my ($vol, $dirs, $file) = splitpath($mapping->path);
	    $key = catdir($dirs);
	} else {
	    $key = $mapping->path;
	}

	unless (exists $groupings{$key}) {
	    $groupings{$key} = Test::MemMonitor::Grouping->new();
	    $groupings{$key}->size(0);
	    $groupings{$key}->name($key);
	}
	
	$groupings{$key}->add_mapping($mapping);
    }
    
    $snapshot->mappings(@mappings);
    foreach my $g (values %groupings) {
	$snapshot->add_grouping($g);
    }

    $snapshot->total_size($total);
    $snapshot->free_size($free);
    
    return $snapshot;
}

sub load_statm {
    my $pid = shift;
    
    my $file = catfile(rootdir(), "proc", $pid, "statm");
    
    my $fh = IO::File->new($file, "r")
	or die "cannot read $file: $!";
    
    my $line = <$fh>;
    
    $fh->close;
    
    my $pagesize = 4096;
    my ($size, $resident, $shared, $text, $lib, $data) = map { $_ * $pagesize } split / /, $line;
    
    
    my $snapshot = Test::MemMonitor::Snapshot->new();
    $snapshot->virtual_size($size);
    $snapshot->resident_size($resident);
    $snapshot->shared_size($shared);
    $snapshot->text_size($text);
    $snapshot->library_size($lib);
    $snapshot->data_size($data);

    return $snapshot;
}

sub load_mappings {
    my $pid = shift;
    
    my $maps = catfile(rootdir(), "proc", $pid, "maps");
    my $smaps = catfile(rootdir(), "proc", $pid, "smaps");
    
    my $fh = IO::File->new($smaps, "r")
	or IO::File->new($maps, "r") or die "cannot read $maps: $!";

# maps has one line per map:
#00400000-004af000 r-xp 00000000 fd:00 18284567                           /bin/bash
# smaps adds in:
#Size:               700 kB
#Rss:                520 kB
#Shared_Clean:       520 kB
#Shared_Dirty:         0 kB
#Private_Clean:        0 kB
#Private_Dirty:        0 kB

    my @mappings;
    while (defined (my $line = <$fh>)) {
	chomp $line;
	if ($line =~ /^\s*([a-f0-9]+)-([a-f0-9]+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)(?:\s+(\S+(?:\s(\(deleted\)))?))?\s*$/) {
	    my $start = $1;
	    my $end = $2;
	    my $permissions = $3;
	    my $offset = $4;
	    my $device = $5;
	    my $inode = $6;
	    my $path = $7;
	    my $deleted = $8;
	    
	    $path = "[anon]" unless $path;
	    
	    my @permissions = split //, $permissions;
	    my @device = split /:/, $device;
	    
	    
	    my $mapping = Test::MemMonitor::Mapping->new();
	    {
		no warnings "portable";
		$mapping->start_address(hex($start));
		$mapping->end_address(hex($end));
	    }
	    $mapping->readable($permissions[0] eq "r" ? 1 : 0);
	    $mapping->writable($permissions[1] eq "w" ? 1 : 0);
	    $mapping->executable($permissions[2] eq "x" ? 1 : 0);
	    $mapping->shared($permissions[3] eq "s" ? 1 : 0);
	    $mapping->private($permissions[3] eq "p" ? 1 : 0);
	    {
		no warnings "portable";
		$mapping->offset(hex($offset));
	    }
	    $mapping->majordev($device[0]);
	    $mapping->minordev($device[1]);
	    $mapping->inode($inode);
	    $mapping->path($path);
	    $mapping->deleted(defined $deleted ? 1 : 0);

	    $mapping->shared_clean(0);
	    $mapping->shared_dirty(0);
	    $mapping->private_clean(0);
	    $mapping->private_dirty(0);
	    
	    push @mappings, $mapping;
	} elsif ($line =~ /^\s*Size\s*:\s*(\d+)\s*kB\s*$/) {
	    # nada
	} elsif ($line =~ /^\s*Rss\s*:\s*(\d+)\s*kB\s*$/) {
	    # nada
	} elsif ($line =~ /^\s*Shared_Clean\s*:\s*(\d+)\s*kB\s*$/) {
	    my $mapping = $mappings[$#mappings];
	    $mapping->shared_clean($1 * 1024);
	} elsif ($line =~ /^\s*Shared_Dirty\s*:\s*(\d+)\s*kB\s*$/) {
	    my $mapping = $mappings[$#mappings];
	    $mapping->shared_dirty($1 * 1024);
	} elsif ($line =~ /^\s*Private_Clean\s*:\s*(\d+)\s*kB\s*$/) {
	    my $mapping = $mappings[$#mappings];
	    $mapping->private_clean($1 * 1024);
	} elsif ($line =~ /^\s*Private_Dirty\s*:\s*(\d+)\s*kB\s*$/) {
	    my $mapping = $mappings[$#mappings];
	    $mapping->private_dirty($1 * 1024);
	} else {
	    warn "malformed data at line $. [$line]\n";
	}
    } 
    $fh->close;

    return @mappings;
}

1;

=pod

=back

=head1 AUTHORS

Daniel P. Berrange <berrange@redhat.com>

=head1 LICENSE / COPYRIGHT

Copyright (C) 2006 Red Hat

This software can be distributed under the terms of the GPL v2 or later

=head1 SEE ALSO

=cut

